use crate::parser::{Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::rc::Rc;

pub struct AlternateParser<P1, P2> {
    pub(crate) first: P1,
    pub(crate) second: P2,
}

impl<'a, TToken, T: 'a, TContext, P1, P2> Parser<'a, TToken, T, TContext>
    for AlternateParser<P1, P2>
where
    TToken: Debug + Clone,
    P1: Parser<'a, TToken, T, TContext>,
    P2: Parser<'a, TToken, T, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        match self.first.parse(context.clone(), input) {
            Err(_) => self.second.parse(context, input),
            r => r,
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::alternate;
    use crate::combinators::ParserExt;
    use crate::parser::parser_test_helper;
    use crate::text::char_parser;

    #[test]
    fn alternate_test() {
        fn parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, char> {
            char_parser('a')
                .alternate(char_parser('b'))
                .alternate(char_parser('c'))
                .parse(context, input)
        }

        parser_test_helper(ParserContext::new_with_str("a", ()), &'a', parser);
        parser_test_helper(ParserContext::new_with_str("b", ()), &'b', parser);
        parser_test_helper(ParserContext::new_with_str("c", ()), &'c', parser);
    }

    #[test]
    fn alternate_macro_test() {
        fn parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, char> {
            alternate!(char_parser('a'), char_parser('b'), char_parser('c')).parse(context, input)
        }

        parser_test_helper(ParserContext::new_with_str("a", ()), &'a', parser);
        parser_test_helper(ParserContext::new_with_str("b", ()), &'b', parser);
        parser_test_helper(ParserContext::new_with_str("c", ()), &'c', parser);
    }

    #[test]
    fn alternate_function_test() {
        fn parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, char> {
            alternate!(char_parser('a'), char_parser('b')).parse(context, input)
        }

        parser_test_helper(ParserContext::new_with_str("a", ()), &'a', |c, i| {
            let parser = alternate!(parser, char_parser('c'));
            let parser = alternate!(parser, char_parser('d'));
            parser.parse(c, i)
        });
    }

    #[test]
    fn alternate_pure_function_test() {
        fn parser1(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, &[char]> {
            alternate!(char_parser('a').literal(), char_parser('b').literal()).parse(context, input)
        }

        fn parser2(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, &[char]> {
            alternate!(char_parser('c').literal(), char_parser('d').literal()).parse(context, input)
        }

        parser_test_helper(
            ParserContext::new_with_str("a", ()),
            &"a".to_owned(),
            |c, i| {
                alternate!(parser1, parser2)
                    .map(|x| x.iter().map(|x| x.clone()).collect())
                    .parse(c, i)
            },
        );
    }
}
