use crate::parser::{FailedParserResult, Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub struct QuoteParser<P, PLeft, PRight, TLeft, TRight> {
    pub(crate) parser: P,
    pub(crate) left: PLeft,
    pub(crate) right: PRight,
    pub(crate) left_data: PhantomData<TLeft>,
    pub(crate) right_data: PhantomData<TRight>,
}

impl<'a, TToken, T: 'a, TLeft: 'a, TRight: 'a, TContext, P, PLeft, PRight>
    Parser<'a, TToken, Vec<T>, TContext> for QuoteParser<P, PLeft, PRight, TLeft, TRight>
where
    TToken: Debug + Clone,
    P: Parser<'a, TToken, T, TContext>,
    PLeft: Parser<'a, TToken, TLeft, TContext>,
    PRight: Parser<'a, TToken, TRight, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<T>> {
        let (mut input, _) = match self.left.parse(context.clone(), input) {
            Err(r) => Err(FailedParserResult::new(
                r.input(),
                format!("Failed to parse left quote: {}", r.message()),
            )),
            r => r,
        }?;

        let mut result = vec![];

        loop {
            if let Ok((i, _)) = self.right.parse(context.clone(), input) {
                input = i;
                break;
            }

            if let Ok((i, r)) = self.parser.parse(context.clone(), input) {
                input = i;
                result.push(r);
            } else {
                return Err(FailedParserResult::new_with_str(
                    input,
                    "Quote without the right quote.",
                ));
            }
        }
        Ok((input, result))
    }
}

pub struct SeparateBy1Parser<P, PSeparator, TSeparator> {
    pub(crate) parser: P,
    pub(crate) separator: PSeparator,
    pub(crate) phantom_data: PhantomData<TSeparator>,
}

impl<'a, TToken, T: 'a, TSeparator: 'a, TContext, P, PSeparator>
    Parser<'a, TToken, Vec<T>, TContext> for SeparateBy1Parser<P, PSeparator, TSeparator>
where
    TToken: Debug + Clone,
    P: Parser<'a, TToken, T, TContext>,
    PSeparator: Parser<'a, TToken, TSeparator, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<T>> {
        let (mut input, first) = self.parser.parse(context.clone(), input)?;
        let mut result = vec![first];

        loop {
            if let Ok((i, _)) = self.separator.parse(context.clone(), input) {
                let (i, r) = self.parser.parse(context.clone(), i)?;
                input = i;
                result.push(r);
            } else {
                break;
            }
        }

        Ok((input, result))
    }
}

pub struct SeparateOrEndBy1Parser<P, PSeparator, TSeparator> {
    pub(crate) parser: P,
    pub(crate) separator: PSeparator,
    pub(crate) phantom_data: PhantomData<TSeparator>,
}

impl<'a, TToken, T: 'a, TSeparator: 'a, TContext, P, PSeparator>
    Parser<'a, TToken, Vec<T>, TContext> for SeparateOrEndBy1Parser<P, PSeparator, TSeparator>
where
    TToken: Debug + Clone,
    P: Parser<'a, TToken, T, TContext>,
    PSeparator: Parser<'a, TToken, TSeparator, TContext>,
{
    fn parse(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<T>> {
        let (mut input, first) = self.parser.parse(context.clone(), input)?;
        let mut result = vec![first];

        loop {
            if let Ok((i, _)) = self.separator.parse(context.clone(), input) {
                if let Ok((i, r)) = self.parser.parse(context.clone(), i) {
                    result.push(r);
                    input = i;
                } else {
                    break;
                }
            } else {
                break;
            }
        }

        Ok((input, result))
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::combinators::{
        end_by, end_by1, quote, separate_by, separate_by1, separate_or_end_by, separate_or_end_by1,
        ParserExt,
    };
    use crate::parser::{failed_parser_test_helper, parser_test_helper, satisfy};
    use crate::text::char_parser;

    #[test]
    fn quote_test() {
        fn parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, Vec<char>> {
            quote(
                char_parser('\''),
                satisfy(|c: &char| c.is_ascii()),
                char_parser('\''),
            )
            .parse(context, input)
        }

        parser_test_helper(
            ParserContext::new_with_str("'abc'", ()),
            &vec!['a', 'b', 'c'],
            parser,
        );
        parser_test_helper(ParserContext::new_with_str("''", ()), &vec![], parser);
        failed_parser_test_helper(ParserContext::new_with_str("asd", ()), parser);

        parser_test_helper(
            ParserContext::new_with_str("'a'b", ()),
            &vec!['a'],
            |c, i| {
                quote(char_parser('\''), char_parser('a'), char_parser('\''))
                    .left(char_parser('b'))
                    .parse(c, i)
            },
        )
    }

    #[test]
    fn separate_by1_test() {
        parser_test_helper(
            ParserContext::new_with_str("a,b,c", ()),
            &vec!['a', 'b', 'c'],
            |c, i| separate_by1(satisfy(|c: &char| c.is_ascii()), char_parser(',')).parse(c, i),
        );
        parser_test_helper(ParserContext::new_with_str("a", ()), &vec!['a'], |c, i| {
            separate_by1(satisfy(|c: &char| c.is_ascii()), char_parser(',')).parse(c, i)
        });
        parser_test_helper(
            ParserContext::new_with_str("bbc", ()),
            &vec!['b'],
            |c, i| separate_by1(satisfy(|c: &char| c.is_ascii()), char_parser(',')).parse(c, i),
        );
    }

    #[test]
    fn separate_by_test() {
        parser_test_helper(
            ParserContext::new_with_str("1,2,3", ()),
            &vec!['1', '2', '3'],
            |c, i| separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')).parse(c, i),
        );
        parser_test_helper(ParserContext::new_with_str("1", ()), &vec!['1'], |c, i| {
            separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')).parse(c, i)
        });
        parser_test_helper(ParserContext::new_with_str("", ()), &vec![], |c, i| {
            separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')).parse(c, i)
        });
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &vec!['a'],
            |c, i| separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')).parse(c, i),
        );
    }

    #[test]
    fn end_by1_test() {
        parser_test_helper(
            ParserContext::new_with_str("aab", ()),
            &vec!['a', 'a'],
            |c, i| end_by1(char_parser('a'), char_parser('b')).parse(c, i),
        );
        parser_test_helper(ParserContext::new_with_str("ab", ()), &vec!['a'], |c, i| {
            end_by1(char_parser('a'), char_parser('b')).parse(c, i)
        });
        failed_parser_test_helper(ParserContext::new_with_str("b", ()), |c, i| {
            end_by1(char_parser('a'), char_parser('b')).parse(c, i)
        });
        failed_parser_test_helper(ParserContext::new_with_str("cd", ()), |c, i| {
            end_by1(char_parser('a'), char_parser('b')).parse(c, i)
        });
    }

    #[test]
    fn end_by_test() {
        parser_test_helper(
            ParserContext::new_with_str("aab", ()),
            &vec!['a', 'a'],
            |c, i| end_by(char_parser('a'), char_parser('b')).parse(c, i),
        );
        parser_test_helper(ParserContext::new_with_str("ab", ()), &vec!['a'], |c, i| {
            end_by(char_parser('a'), char_parser('b')).parse(c, i)
        });
        parser_test_helper(ParserContext::new_with_str("b", ()), &vec![], |c, i| {
            end_by(char_parser('a'), char_parser('b')).parse(c, i)
        });
        failed_parser_test_helper(ParserContext::new_with_str("cd", ()), |c, i| {
            end_by(char_parser('a'), char_parser('b')).parse(c, i)
        });
    }

    #[test]
    fn separate_or_end_by1_test() {
        parser_test_helper(
            ParserContext::new_with_str("1,2,3,", ()),
            &vec!['1', '2', '3'],
            |c, i| {
                separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(','))
                    .parse(c, i)
            },
        );
        parser_test_helper(ParserContext::new_with_str("1,", ()), &vec!['1'], |c, i| {
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
        });
        parser_test_helper(
            ParserContext::new_with_str("1,2,3", ()),
            &vec!['1', '2', '3'],
            |c, i| {
                separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(','))
                    .parse(c, i)
            },
        );
        parser_test_helper(ParserContext::new_with_str("1", ()), &vec!['1'], |c, i| {
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
        });
        failed_parser_test_helper(ParserContext::new_with_str("abc", ()), |c, i| {
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
        });
    }

    #[test]
    fn separate_or_end_by_test() {
        parser_test_helper(
            ParserContext::new_with_str("1,2,3,", ()),
            &vec!['1', '2', '3'],
            |c, i| {
                separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
            },
        );
        parser_test_helper(ParserContext::new_with_str("1,", ()), &vec!['1'], |c, i| {
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
        });
        parser_test_helper(
            ParserContext::new_with_str("1,2,3", ()),
            &vec!['1', '2', '3'],
            |c, i| {
                separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
            },
        );
        parser_test_helper(ParserContext::new_with_str("1", ()), &vec!['1'], |c, i| {
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
        });
        parser_test_helper(ParserContext::new_with_str("abc", ()), &vec![], |c, i| {
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')).parse(c, i)
        });
    }
}
