use crate::parser::{FailedParserResult, Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub struct SucceedParser<T> {
    pub(crate) value: T,
}

impl<'a, TToken, T: 'a, TContext> Parser<'a, TToken, T, TContext> for SucceedParser<T>
where
    TToken: Debug + Clone,
    T: Debug + Clone,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        Ok((input, self.value.clone()))
    }
}

pub struct FailParser<T> {
    pub(crate) phantom_data: PhantomData<T>,
}

impl<'a, TToken, T: 'a, TContext> Parser<'a, TToken, T, TContext> for FailParser<T>
where
    TToken: Debug + Clone,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        Err(FailedParserResult::new_with_str(
            input,
            "Default failed parser.",
        ))
    }
}

pub struct FailWithMessageParser<T> {
    pub(crate) message: String,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<'a, TToken, T: 'a, TContext> Parser<'a, TToken, T, TContext> for FailWithMessageParser<T>
where
    TToken: Debug + Clone,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        Err(FailedParserResult::new(input, self.message.clone()))
    }
}

pub struct SatisfyParser<F> {
    pub(crate) predicate: F,
}

impl<'a, TToken, TContext, F> Parser<'a, TToken, TToken, TContext> for SatisfyParser<F>
where
    TToken: Debug + Clone + 'a,
    F: Fn(&TToken) -> bool,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TToken> {
        if input.is_empty() {
            return Err(FailedParserResult::new_with_str(input, "Input is empty"));
        }

        if (self.predicate)(&input[0]) {
            Ok((&input[1..], input[0].clone()))
        } else {
            Err(FailedParserResult::new_with_str(input, "Predicate failed."))
        }
    }
}

pub struct SatifiedMapParser<F> {
    pub(crate) mapper: F,
}

impl<'a, TToken, TContext, T: 'a, F> Parser<'a, TToken, T, TContext> for SatifiedMapParser<F>
where
    TToken: Debug + Clone,
    F: Fn(&TToken) -> Option<T>,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        if input.is_empty() {
            return Err(FailedParserResult::new_with_str(input, "Input is empty"));
        }

        if let Some(result) = (self.mapper)(&input[0]) {
            Ok((&input[1..], result))
        } else {
            Err(FailedParserResult::new_with_str(input, "Predicate failed."))
        }
    }
}

pub struct AnyParser {}

impl<'a, TToken, TContext> Parser<'a, TToken, TToken, TContext> for AnyParser
where
    TToken: Debug + Clone + 'a,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TToken> {
        if input.is_empty() {
            Err(FailedParserResult::new_with_str(input, "Input is empty"))
        } else {
            Ok((&input[1..], input[0].clone()))
        }
    }
}

pub struct TakeParser {
    pub(crate) count: usize,
}

impl<'a, TToken, TContext> Parser<'a, TToken, Vec<TToken>, TContext> for TakeParser
where
    TToken: Debug + Clone + 'a,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<TToken>> {
        if input.len() < self.count {
            Err(FailedParserResult::new(
                input,
                format!(
                    "The input doesn't contain enough {} elemements.",
                    self.count
                ),
            ))
        } else {
            Ok((
                &input[self.count..],
                (&input[..self.count]).iter().map(|x| x.clone()).collect(),
            ))
        }
    }
}

pub struct SkipParser {
    pub(crate) count: usize,
}

impl<'a, TToken, TContext> Parser<'a, TToken, (), TContext> for SkipParser
where
    TToken: Debug + Clone,
{
    fn parse(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, ()> {
        if input.len() < self.count {
            Err(FailedParserResult::new(
                input,
                format!(
                    "The input doesn't contain enough {} elemements.",
                    self.count
                ),
            ))
        } else {
            Ok((&input[self.count..], ()))
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{
        any, fail, fail_with_message, failed_parser_test_helper, parser_test_helper, satified_map,
        satisfy, skip, succeed, take,
    };

    #[test]
    fn succeed_test() {
        parser_test_helper(ParserContext::new_with_str("abc", ()), &(), |c, i| {
            succeed(()).parse(c, i)
        });
    }

    #[test]
    fn fail_test() {
        failed_parser_test_helper(ParserContext::new_with_str("abc", ()), |c, i| {
            fail::<char, (), ()>().parse(c, i)
        });
    }

    #[test]
    fn fail_with_message_test() {
        failed_parser_test_helper(ParserContext::new_with_str("abc", ()), |c, i| {
            fail_with_message::<char, (), ()>("Failed!").parse(c, i)
        });
    }

    #[test]
    fn satisfy_test() {
        parser_test_helper(ParserContext::new_with_str("abc", ()), &'a', |c, i| {
            satisfy(|x| x == &'a').parse(c, i)
        });

        failed_parser_test_helper(ParserContext::new_with_str("abc", ()), |c, i| {
            satisfy(|x| x == &'b').parse(c, i)
        });
    }

    #[test]
    fn satified_map_test() {
        parser_test_helper(ParserContext::new_with_str("123", ()), &1, |c, i| {
            satified_map(|x: &char| x.to_digit(10)).parse(c, i)
        });

        parser_test_helper(ParserContext::new_with_str("23", ()), &2, |c, i| {
            satified_map(|x: &char| x.to_digit(10)).parse(c, i)
        });

        parser_test_helper(ParserContext::new_with_str("3", ()), &3, |c, i| {
            satified_map(|x: &char| x.to_digit(10)).parse(c, i)
        });
    }

    #[test]
    fn any_test() {
        parser_test_helper(ParserContext::new_with_str("abc", ()), &'a', |c, i| {
            any().parse(c, i)
        });

        parser_test_helper(ParserContext::new_with_str("cde", ()), &'c', |c, i| {
            any().parse(c, i)
        });
    }

    #[test]
    fn take_test() {
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &("hello".chars().collect()),
            |c, i| take(5).parse(c, i),
        );

        failed_parser_test_helper(ParserContext::new_with_str("abcd", ()), |c, i| {
            take(5).parse(c, i)
        });
    }

    #[test]
    fn skip_test() {
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &(),
            |c, i| skip(5).parse(c, i),
        );

        failed_parser_test_helper(ParserContext::new_with_str("abcd", ()), |c, i| {
            skip(5).parse(c, i)
        });
    }
}
