mod string_parsers;

use crate::parser::{satisfy, Parser};
use crate::text::string_parsers::StringParser;

pub fn char_parser<'a, TContext>(c: char) -> impl Parser<'a, char, char, TContext> {
    satisfy(move |x| *x == c)
}

pub fn string_parser<'a, TContext>(
    str: &'static str,
) -> impl Parser<'a, char, &'a [char], TContext> {
    StringParser { str }
}

pub fn one_of<'a, TContext>(str: &'static str) -> impl Parser<'a, char, char, TContext> {
    let str: Vec<char> = str.chars().collect();
    satisfy(move |c: &char| str.contains(c))
}

pub fn char_satisfy<'a, TContext, F>(condition: F) -> impl Parser<'a, char, char, TContext>
where
    F: Fn(&char) -> bool,
{
    satisfy(move |x: &char| condition(x))
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{
        failed_parser_test_helper, parser_test_helper, ParserContext, ParserResult,
    };
    use std::cell::RefCell;
    use std::rc::Rc;

    #[test]
    fn char_parser_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'a',
            |context, input| char_parser('a').parse(context, input),
        );

        failed_parser_test_helper(ParserContext::new_with_str("bc", ()), |context, input| {
            char_parser('a').parse(context, input)
        });
    }

    #[test]
    fn string_parser_test() {
        parser_test_helper(
            ParserContext::new_with_str("Hello, world!", ()),
            &"Hello".to_owned(),
            |context, input| {
                string_parser("Hello")
                    .map(|x| x.iter().collect())
                    .parse(context, input)
            },
        );

        fn test_parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, String> {
            let (input, first) = string_parser("hello, ").parse(context.clone(), input)?;
            let (input, second) = string_parser("world!").parse(context.clone(), input)?;

            Ok((
                input,
                first.iter().collect::<String>() + second.iter().collect::<String>().as_str(),
            ))
        }

        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &"hello, world!".to_owned(),
            test_parser,
        );
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &(),
            |context, input| test_parser.map(|_| ()).parse(context, input),
        )
    }

    #[test]
    fn char_satsify_test() {
        parser_test_helper(ParserContext::new_with_str("abc", ()), &'a', |c, i| {
            char_satisfy(char::is_ascii).parse(c, i)
        });
    }
}
